@@ -90,7 +90,7 @@ static void test_select( IWbemServices *services )
         L"SELECT\na\rFROM\tb",
         L"SELECT * FROM Win32_Process WHERE Caption LIKE \"%firefox.exe\"",
         L"SELECT * FROM Win32_VideoController where availability = '3'",
-        L"SELECT * FROM Win3_BIOS WHERE NAME <> NULL",
+        L"SELECT * FROM Win32_BIOS WHERE NAME <> NULL",
         L"SELECT * FROM Win32_BIOS WHERE NULL = NAME",
         L"SELECT * FROM Win32_LogicalDiskToPartition",
         L"SELECT * FROM Win32_DiskDriveToDiskPartition",
@@ -140,6 +140,82 @@ static void test_select( IWbemServices *services )
     SysFreeString( query );
 }
 
+static void check_explorer_like_query( IWbemServices *services, const WCHAR *str, BOOL expect_success)
+{
+    HRESULT hr;
+    IWbemClassObject *obj[2];
+    BSTR wql = SysAllocString( L"wql" ), query = SysAllocString( str );
+    LONG flags = WBEM_FLAG_RETURN_IMMEDIATELY | WBEM_FLAG_FORWARD_ONLY;
+    ULONG count;
+    IEnumWbemClassObject *result;
+
+    hr = IWbemServices_ExecQuery( services, wql, query, flags, NULL, &result );
+    if (hr == S_OK)
+    {
+        VARIANT var;
+        IEnumWbemClassObject_Next( result, 10000, 2, obj, &count );
+
+        ok( count == (expect_success ? 1 : 0), "expected to get %d results but got %lu\n",
+                (expect_success ? 1 : 0), count);
+
+        if (count)
+        {
+            BSTR caption;
+            hr = IWbemClassObject_Get( obj[0], L"Caption", 0, &var, NULL, NULL );
+            ok( hr == WBEM_S_NO_ERROR, "IWbemClassObject_Get failed %#lx", hr);
+            caption = V_BSTR(&var);
+            ok( !wcscmp( caption, L"explorer.exe" ), "%s is not explorer.exe\n", debugstr_w(caption));
+            VariantClear( &var );
+        }
+
+        while (count--)
+            IWbemClassObject_Release( obj[count] );
+    }
+
+    SysFreeString( wql );
+    SysFreeString( query );
+}
+
+
+static void test_like_query( IWbemServices *services )
+{
+    int i;
+    WCHAR query[250];
+
+    struct {
+        BOOL expect_success;
+        const WCHAR *str;
+    } queries[] = {
+        { TRUE,  L"explorer%" },
+        { FALSE, L"xplorer.exe" },
+        { FALSE, L"explorer.ex" },
+        { TRUE,  L"%explorer%" },
+        { TRUE,  L"explorer.exe%" },
+        { TRUE,  L"%explorer.exe%" },
+        { TRUE,  L"%plorer.exe" },
+        { TRUE,  L"%plorer.exe%" },
+        { TRUE,  L"__plorer.exe" },
+        { TRUE,  L"e_plorer.exe" },
+        { FALSE, L"_plorer.exe" },
+        { TRUE,  L"%%%plorer.e%" },
+        { TRUE,  L"%plorer.e%" },
+        { TRUE,  L"%plorer.e_e" },
+        { TRUE,  L"%plorer.e_e" },
+        { TRUE,  L"explore%exe" },
+        { FALSE, L"fancy_explore.exe" },
+        { FALSE, L"fancy%xplore%exe" },
+        { FALSE, L"%%%f%xplore%exe" },
+    };
+
+    for (i = 0; i < ARRAYSIZE(queries); i++)
+    {
+        wsprintfW( query, L"SELECT * FROM Win32_Process WHERE Caption LIKE '%ls'", queries[i].str );
+        trace("%s\n", wine_dbgstr_w(query));
+        check_explorer_like_query( services, query, queries[i].expect_success );
+    }
+}
+
+
 static void test_associators( IWbemServices *services )
 {
     static const WCHAR *test[] =
@@ -407,6 +483,7 @@ static void test_Win32_Bios( IWbemServices *services )
     check_property( obj, L"SMBIOSBIOSVersion", VT_BSTR, CIM_STRING );
     check_property( obj, L"SMBIOSMajorVersion", VT_I4, CIM_UINT16 );
     check_property( obj, L"SMBIOSMinorVersion", VT_I4, CIM_UINT16 );
+    check_property( obj, L"Status", VT_BSTR, CIM_STRING );
     check_property( obj, L"Version", VT_BSTR, CIM_STRING );
 
     IWbemClassObject_Release( obj );
@@ -472,6 +549,7 @@ static void test_Win32_Process( IWbemServices *services, BOOL use_full_path )
     IWbemClassObject *process, *sig_in, *sig_out, *out, *params;
     WCHAR cmdlineW[MAX_PATH + 64 + 1];
     IWbemQualifierSet *qualifiers;
+    WCHAR executable_path[255];
     VARIANT retval, val;
     SAFEARRAY *names;
     LONG bound, i;
@@ -641,7 +719,6 @@ static void test_Win32_Process( IWbemServices *services, BOOL use_full_path )
     hr = IWbemServices_ExecMethod( services, class, method, 0, NULL, NULL, &out, NULL );
     ok( hr == S_OK, "failed to execute method %#lx\n", hr );
     SysFreeString( method );
-    SysFreeString( class );
 
     type = 0xdeadbeef;
     VariantInit( &retval );
@@ -688,8 +765,25 @@ static void test_Win32_Process( IWbemServices *services, BOOL use_full_path )
     hr = IWbemQualifierSet_Get( qualifiers, L"ID", 0, &val, &flavor );
     ok( hr == WBEM_E_NOT_FOUND, "got %#lx\n", hr );
 
+    /* Test instance properties */
+
+    hr = IWbemServices_GetObject( services, class, 0, NULL, &process, NULL );
+    ok( hr == S_OK, "got %#lx\n", hr );
+
+    type = 0xdeadbeef;
+    VariantInit( &val );
+    hr = IWbemClassObject_Get( process, L"ExecutablePath", 0, &val, &type, NULL );
+    ok( hr == S_OK, "IWbemClassObject_Get failed with %#lx\n", hr );
+    ok( V_VT( &val ) == VT_BSTR, "unexpected variant type 0x%x\n", V_VT( &val ) );
+    ok( type == CIM_STRING, "unexpected type %#lx\n", type );
+    GetModuleFileNameW( NULL, executable_path, ARRAY_SIZE(executable_path) );
+    ok( !lstrcmpiW( V_BSTR( &val ), executable_path ), "got %s, expected %s\n", wine_dbgstr_w(V_BSTR(&val)), wine_dbgstr_w(executable_path) );
+    VariantClear( &val );
+    IWbemClassObject_Release( process );
+
     IWbemQualifierSet_Release( qualifiers );
     IWbemClassObject_Release( out );
+    SysFreeString( class );
 }
 
 static void test_Win32_ComputerSystem( IWbemServices *services )
@@ -1607,6 +1701,15 @@ static void test_Win32_PhysicalMemory( IWbemServices *services )
             trace( "ConfiguredClockSpeed %ld\n", V_I4( &val ) );
         }
 
+        type = 0xdeadbeef;
+        VariantInit( &val );
+        hr = IWbemClassObject_Get( obj, L"Manufacturer", 0, &val, &type, NULL );
+        ok( hr == S_OK, "got %#lx\n", hr );
+        ok( V_VT( &val ) == VT_BSTR || V_VT( &val ) == VT_NULL, "unexpected variant type %#x\n", V_VT( &val ) );
+        ok( type == CIM_STRING, "unexpected type %#lx\n", type );
+        trace( "Manufacturer %s\n", wine_dbgstr_w(V_BSTR( &val )) );
+        VariantClear( &val );
+
         type = 0xdeadbeef;
         VariantInit( &val );
         hr = IWbemClassObject_Get( obj, L"PartNumber", 0, &val, &type, NULL );
@@ -1633,6 +1736,53 @@ static void test_Win32_PhysicalMemory( IWbemServices *services )
     SysFreeString( wql );
 }
 
+static void test_Win32_PhysicalMemoryArray( IWbemServices *services )
+{
+    BSTR wql = SysAllocString( L"wql" ), query = SysAllocString( L"SELECT * FROM Win32_PhysicalMemoryArray" );
+    IEnumWbemClassObject *result;
+    IWbemClassObject *obj;
+    CIMTYPE type;
+    VARIANT val;
+    DWORD count;
+    HRESULT hr;
+
+    hr = IWbemServices_ExecQuery( services, wql, query, 0, NULL, &result );
+    if (hr != S_OK)
+    {
+        win_skip( "Win32_PhysicalMemoryArray not available\n" );
+        return;
+    }
+
+    for (;;)
+    {
+        hr = IEnumWbemClassObject_Next( result, 10000, 1, &obj, &count );
+        if (hr != S_OK) break;
+
+        check_property( obj, L"Caption", VT_BSTR, CIM_STRING );
+        check_property( obj, L"MaxCapacity", VT_I4, CIM_UINT32 );
+        check_property( obj, L"Model", VT_NULL, CIM_STRING );
+        check_property( obj, L"Name", VT_BSTR, CIM_STRING );
+        check_property( obj, L"Status", VT_NULL, CIM_STRING );
+
+        type = 0xdeadbeef;
+        VariantInit( &val );
+        hr = IWbemClassObject_Get( obj, L"MaxCapacityEx", 0, &val, &type, NULL );
+        ok( hr == S_OK || broken(hr == WBEM_E_NOT_FOUND) /* < win10 */, "got %#lx\n", hr );
+        if (hr == S_OK)
+        {
+            ok( V_VT( &val ) == VT_BSTR, "unexpected variant type %#x\n", V_VT( &val ) );
+            ok( type == CIM_UINT64, "unexpected type %#lx\n", type );
+            VariantClear( &val );
+        }
+
+        IWbemClassObject_Release( obj );
+    }
+
+    IEnumWbemClassObject_Release( result );
+    SysFreeString( query );
+    SysFreeString( wql );
+}
+
 static void test_Win32_IP4RouteTable( IWbemServices *services )
 {
     BSTR wql = SysAllocString( L"wql" ), query = SysAllocString( L"SELECT * FROM Win32_IP4RouteTable" );
@@ -1755,8 +1905,28 @@ static void test_Win32_VideoController( IWbemServices *services )
         check_property( obj, L"__SERVER", VT_BSTR, CIM_STRING );
         check_property( obj, L"AdapterCompatibility", VT_BSTR, CIM_STRING );
         check_property( obj, L"Availability", VT_I4, CIM_UINT16 );
+        check_property( obj, L"Caption", VT_BSTR, CIM_STRING );
         check_property( obj, L"ConfigManagerErrorCode", VT_I4, CIM_UINT32 );
         check_property( obj, L"DriverDate", VT_BSTR, CIM_DATETIME );
+        check_property( obj, L"DriverVersion", VT_BSTR, CIM_STRING );
+        check_property( obj, L"Status", VT_BSTR, CIM_STRING );
+
+        type = 0xdeadbeef;
+        VariantInit( &val );
+        hr = IWbemClassObject_Get( obj, L"AdapterDACType", 0, &val, &type, NULL );
+        ok( hr == S_OK, "got %#lx\n", hr );
+        ok( V_VT( &val ) == VT_BSTR || V_VT( &val ) == VT_NULL, "unexpected variant type 0x%x\n", V_VT( &val ) );
+        ok( type == CIM_STRING, "unexpected type %#lx\n", type );
+        trace( "AdapterDACType %s\n", wine_dbgstr_w(V_BSTR( &val )) );
+        VariantClear( &val );
+
+        type = 0xdeadbeef;
+        VariantInit( &val );
+        hr = IWbemClassObject_Get( obj, L"AdapterRAM", 0, &val, &type, NULL );
+        ok( hr == S_OK, "got %#lx\n", hr );
+        ok( V_VT( &val ) == VT_I4 || V_VT( &val ) == VT_NULL, "unexpected variant type 0x%x\n", V_VT( &val ) );
+        ok( type == CIM_UINT32, "unexpected type %#lx\n", type );
+        trace( "AdapterRAM %lu\n", V_UI4( &val ) );
 
         type = 0xdeadbeef;
         VariantInit( &val );
@@ -1767,7 +1937,6 @@ static void test_Win32_VideoController( IWbemServices *services )
         trace( "installeddisplaydrivers %s\n", wine_dbgstr_w(V_BSTR( &val )) );
         VariantClear( &val );
 
-        check_property( obj, L"Status", VT_BSTR, CIM_STRING );
         IWbemClassObject_Release( obj );
     }
 
@@ -2376,6 +2545,7 @@ START_TEST(query)
     test_query_async( services );
     test_query_semisync( services );
     test_select( services );
+    test_like_query( services );
 
     /* classes */
     test_SoftwareLicensingProduct( services );
@@ -2395,6 +2565,7 @@ START_TEST(query)
     test_Win32_OperatingSystem( services );
     test_Win32_PageFileUsage( services );
     test_Win32_PhysicalMemory( services );
+    test_Win32_PhysicalMemoryArray( services );
     test_Win32_PnPEntity( services );
     test_Win32_Printer( services );
     test_Win32_Process( services, FALSE );
